/*
Copyright (C) 2008  Damien Thebault

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "libasyncns.h"
#include "addrinfoquery.h"
#include "query.h"

#include <structmember.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

static int AddrInfoQuery_init(AddrInfoQuery *self, PyObject *args, PyObject *kwds)
{
	Query_type.tp_init((PyObject*)self, args, kwds);
	/* exception is set by Query.__init__() */
	return -1;
}

static PyObject *AddrInfoQuery_get_done(AddrInfoQuery *self)
{
	PyObject *out = NULL;
	int r;
	struct addrinfo *ai, *i;
	if(self->query.result != NULL)
	{
		out = self->query.result;
		Py_INCREF(out);
		goto get_done_end;
	}
	r = asyncns_getaddrinfo_done(self->query.asyncns->asyncns,
	                             self->query.asyncns_query, &ai);
	asyncns_del_query(self->query.asyncns, (Query*)self);
	if(r != 0)
	{
		libasyncns_set_gaierror(r);
		out = NULL;
		goto get_done_end;
	}
	out = PyList_New(0);
	for (i = ai; i; i = i->ai_next)
	{
		char t[256];
		const char *p = NULL;
		PyObject *tu = PyTuple_New(5);
		PyObject *sockaddr = NULL;
		
		if (i->ai_family == PF_INET)
		{
			sockaddr = PyTuple_New(2);
			struct sockaddr_in *si = (struct sockaddr_in*)i->ai_addr;
			p = inet_ntop(AF_INET, &si->sin_addr, t, sizeof(t));
#if PY_MAJOR_VERSION >= 3
			PyTuple_SetItem(sockaddr, 0, PyUnicode_FromString(p));
			PyTuple_SetItem(sockaddr, 1, PyLong_FromLong(ntohs(si->sin_port)));
#else /* PY_MAJOR_VERSION >= 3 */
			PyTuple_SetItem(sockaddr, 0, PyString_FromString(p));
			PyTuple_SetItem(sockaddr, 1, PyInt_FromLong(ntohs(si->sin_port)));
#endif /* PY_MAJOR_VERSION >= 3 */
		}
		else if (i->ai_family == PF_INET6)
		{
			sockaddr = PyTuple_New(4);
			struct sockaddr_in6 *s6 = (struct sockaddr_in6*)i->ai_addr;
			p = inet_ntop(AF_INET6, &s6->sin6_addr, t, sizeof(t));
#if PY_MAJOR_VERSION >= 3
			PyTuple_SetItem(sockaddr, 0, PyUnicode_FromString(p));
			PyTuple_SetItem(sockaddr, 1, PyLong_FromLong(ntohs(s6->sin6_port)));
			PyTuple_SetItem(sockaddr, 2, PyLong_FromLong(ntohl(s6->sin6_flowinfo)));
			PyTuple_SetItem(sockaddr, 3, PyLong_FromLong(ntohl(s6->sin6_scope_id)));
#else /* PY_MAJOR_VERSION >= 3 */
			PyTuple_SetItem(sockaddr, 0, PyString_FromString(p));
			PyTuple_SetItem(sockaddr, 1, PyInt_FromLong(ntohs(s6->sin6_port)));
			PyTuple_SetItem(sockaddr, 2, PyInt_FromLong(ntohl(s6->sin6_flowinfo)));
			PyTuple_SetItem(sockaddr, 3, PyInt_FromLong(ntohl(s6->sin6_scope_id)));
#endif /* PY_MAJOR_VERSION >= 3 */
		}
#if PY_MAJOR_VERSION >= 3
		PyTuple_SetItem(tu, 0, PyLong_FromLong(i->ai_family));
		PyTuple_SetItem(tu, 1, PyLong_FromLong(i->ai_socktype));
		PyTuple_SetItem(tu, 2, PyLong_FromLong(i->ai_protocol));
		PyTuple_SetItem(tu, 3, PyUnicode_FromString(i->ai_canonname?i->ai_canonname:""));
#else /* PY_MAJOR_VERSION >= 3 */
		PyTuple_SetItem(tu, 0, PyInt_FromLong(i->ai_family));
		PyTuple_SetItem(tu, 1, PyInt_FromLong(i->ai_socktype));
		PyTuple_SetItem(tu, 2, PyInt_FromLong(i->ai_protocol));
		PyTuple_SetItem(tu, 3, PyString_FromString(i->ai_canonname?i->ai_canonname:""));
#endif /* PY_MAJOR_VERSION >= 3 */
		PyTuple_SetItem(tu, 4, sockaddr);
		PyList_Append(out, tu);
	}
	Query_store_result(&(self->query), out);
	asyncns_freeaddrinfo(ai);
get_done_end:
	self->query.asyncns_query = NULL;
	return out;
}

static PyMethodDef AddrInfoQuery_methods[] =
{
	{ .ml_name = "get_done",
	  .ml_meth = (PyCFunction)AddrInfoQuery_get_done,
	  .ml_flags = METH_NOARGS,
	  .ml_doc = "get_done()\n"
	            "Return the result of the query.\n"
	            "The result is a list of tuples:\n"
	            "(host, port) for IPv4, "
	            "(host, port, flowinfo, scopeid) for IPv6." },
	{ NULL, NULL, 0, NULL },
};

static PyMemberDef AddrInfoQuery_members[] =
{
	{ NULL, 0, 0, 0, NULL },
};


PyTypeObject AddrInfoQuery_type =
{
#if PY_MAJOR_VERSION >= 3
	PyVarObject_HEAD_INIT(NULL,0)
#else /* PY_MAJOR_VERSION >= 3 */
	PyObject_HEAD_INIT(NULL)
#endif /* PY_MAJOR_VERSION >= 3 */

	.tp_name = "libasyncns.AddrInfoQuery",
	.tp_doc  = "The class for getaddrinfo() calls.",

	.tp_init    = (initproc)  AddrInfoQuery_init, /* no init, see Asyncns */

	.tp_methods = AddrInfoQuery_methods,
	.tp_members = AddrInfoQuery_members,

	.tp_flags     = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
	.tp_basicsize = sizeof(AddrInfoQuery),
};

void initAddrInfoQuery(PyObject *libasyncns_module)
{
	AddrInfoQuery_type.tp_base = &Query_type; /* windows workaround */
	PyType_Ready(&AddrInfoQuery_type);
	Py_INCREF(&AddrInfoQuery_type);
	PyModule_AddObject(libasyncns_module, "AddrInfoQuery", (PyObject*)&AddrInfoQuery_type);
}
