/*
Copyright (C) 2008  Damien Thebault

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "asyncns.h"
#include "addrinfoquery.h"
#include "nameinfoquery.h"
#include "resquery.h"

#include <structmember.h>

#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <netdb.h>

static Asyncns *Asyncns_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
	Asyncns *self = (Asyncns*)type->tp_alloc(type, 0);
	self->asyncns = NULL;
	self->queries = NULL;
	self->nqueries = 0;
	return self;
}

static int Asyncns_init(Asyncns *self, PyObject *args, PyObject *kwds)
{
	int n_proc = 5;
	static char *kwlist[] = {"n_proc", NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|i", kwlist, &n_proc))
		return -1;
	self->asyncns = asyncns_new(n_proc);
	return 0;
}

static void Asyncns_dealloc(Asyncns *self)
{
	if(self->asyncns != NULL)
	{
		asyncns_free(self->asyncns);
		self->asyncns = NULL;
	}
}

static Query *_asyncns_get_query(Asyncns *asyncns,
                                 asyncns_query_t *asyncns_query)
{
	int i;
	for(i = 0 ; i<asyncns->nqueries ; i++)
	{
		if(asyncns->queries[i]->asyncns_query == asyncns_query)
		{
			return asyncns->queries[i];
		}
	}
	return NULL;
}

int asyncns_add_query(Asyncns *asyncns, Query *query)
{
	Query **queries;
	queries = (Query **)realloc(asyncns->queries, (asyncns->nqueries+1)*
	                            sizeof(*(asyncns->queries)));
	if(queries == NULL)
	{
		return -1;
	}
	asyncns->nqueries++;
	asyncns->queries = queries;
	asyncns->queries[asyncns->nqueries-1] = query;
	Py_INCREF(query);
	return 0;
}

int asyncns_del_query(Asyncns *asyncns, Query *query)
{
	int i, j;
	unsigned char found = 0;
	for(i=0 ; i<asyncns->nqueries ; i++)
	{
		if(asyncns->queries[i] == query)
		{
			found = 1;
			for(j=i ; j<asyncns->nqueries - 1 ; j++)
			{
				asyncns->queries[j] = asyncns->queries[j+1];
			}
			asyncns->nqueries--;
		}
	}
	if(found)
	{
		asyncns->queries = (Query **)realloc(asyncns->queries, asyncns->nqueries
		                                     * sizeof(*(asyncns->queries)));
		Py_DECREF(query);
	}
	return found?0:-1;
}

static AddrInfoQuery *Asyncns_getaddrinfo(Asyncns *self, PyObject *args, PyObject *kwds)
{
	char *host = NULL, *port = NULL;
	int family = AF_UNSPEC;
	static char *kwlist[] = {"host", "port", "family", "socktype", "proto", "flags", NULL};
	struct addrinfo hints;
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = 0;
	hints.ai_protocol = 0;
	hints.ai_flags    = 0;
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "zz|iiii", kwlist, &host, &port,
	                                &hints.ai_family, &hints.ai_socktype,
	                                &hints.ai_protocol, &hints.ai_flags))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
	AddrInfoQuery *q = (AddrInfoQuery *)PyObject_MyNew(AddrInfoQuery_type);
	Query_store_asyncns(&q->query, self,
	                    asyncns_getaddrinfo(self->asyncns, host, port, &hints));
	asyncns_add_query(self, (Query*)q);
	return q;
}

/* return the port number, even if the argument is a service name */
static int get_port(PyObject *port)
{
#if PY_MAJOR_VERSION >= 3
	if(PyUnicode_Check(port))
#else /* PY_MAJOR_VERSION >= 3 */
	if(PyString_Check(port))
#endif /* PY_MAJOR_VERSION >= 3 */
	{
#if PY_MAJOR_VERSION >= 3
		PyObject *encport;
		encport = PyUnicode_AsUTF8String(port);
		char *serv = PyBytes_AsString(encport);
#else /* PY_MAJOR_VERSION >= 3 */
		char *serv = PyString_AsString(port);
#endif /* PY_MAJOR_VERSION >= 3 */
		struct servent *se = getservbyname(serv, NULL);
#if PY_MAJOR_VERSION >= 3
		Py_DECREF(encport);
#endif /* PY_MAJOR_VERSION >= 3 */
		if(se == NULL)
		{
			PyErr_Format(PyExc_AttributeError, "unknown port \"%s\"", serv);
			return -1;
		}
		return ntohs(se->s_port);
	}
#if PY_MAJOR_VERSION >= 3
	else if(PyLong_Check(port))
	{
		return PyLong_AsLong(port);
	}
#else /* PY_MAJOR_VERSION >= 3 */
	else if(PyInt_Check(port))
	{
		return PyInt_AsLong(port);
	}
#endif /* PY_MAJOR_VERSION >= 3 */
	else
	{
		return -2;
	}
}

static NameInfoQuery *Asyncns_getnameinfo(Asyncns *self, PyObject *args, PyObject *kwds)
{
	PyObject *tuple = NULL;
	int flags = 0;
	static char *kwlist[] = {"sockaddr", "flags", NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "O!|i", kwlist, &PyTuple_Type, &tuple, &flags))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
	
	NameInfoQuery *q = (NameInfoQuery *)PyObject_MyNew(NameInfoQuery_type);
	
	struct sockaddr *sa;
	int sa_size = 0;
	
	struct sockaddr_in si;
	struct sockaddr_in6 s6;
	
	switch(PyTuple_Size(tuple))
	{
		case 2: /* ipv4 */
			{
				memset(&si, 0, sizeof(si));
				si.sin_family = AF_INET;
				
				PyObject *host = PyTuple_GetItem(tuple, 0);
				PyObject *port = PyTuple_GetItem(tuple, 1);
				
#if PY_MAJOR_VERSION >= 3
				if(!PyUnicode_Check(host))
#else /* PY_MAJOR_VERSION >= 3 */
				if(!PyString_Check(host))
#endif /* PY_MAJOR_VERSION >= 3 */
				{
					PyErr_SetString(PyExc_TypeError, "first item of sockaddr must be a string");
					return NULL;
				}
#if PY_MAJOR_VERSION >= 3
				if(!(PyUnicode_Check(port) || PyLong_Check(port)))
#else /* PY_MAJOR_VERSION >= 3 */
				if(!(PyString_Check(port) || PyInt_Check(port)))
#endif /* PY_MAJOR_VERSION >= 3 */
				{
					PyErr_SetString(PyExc_AttributeError, "second item of sockaddr must be an int or a string");
					return NULL;
				}
				
#if PY_MAJOR_VERSION >= 3
				PyObject *enchost;
				enchost = PyUnicode_AsUTF8String(host);
				char *addr = PyBytes_AsString(enchost);
#else /* PY_MAJOR_VERSION >= 3 */
				char *addr = PyString_AsString(host);
#endif /* PY_MAJOR_VERSION >= 3 */
				int rc = inet_pton(AF_INET, addr, &si.sin_addr);
#if PY_MAJOR_VERSION >= 3
				Py_DECREF(enchost);
#endif /* PY_MAJOR_VERSION >= 3 */
				if(rc < 0)
				{
					PyErr_SetFromErrno(PyExc_AttributeError);
					return NULL;
				}
				
				int port_number = get_port(port);
				if(port_number < 0)
				{
					/* exception is set by get_port() */
					return NULL;
				}
				si.sin_port = htons(port_number);
				
				sa = (struct sockaddr*)&si;
				sa_size = sizeof(si);
			}
			break;
		case 4: /* ipv6 */
			{
				memset(&s6, 0, sizeof(s6));
				s6.sin6_family = AF_INET6;
				
				PyObject *host = PyTuple_GetItem(tuple, 0);
				PyObject *port = PyTuple_GetItem(tuple, 1);
				PyObject *flowinfo = PyTuple_GetItem(tuple, 2);
				PyObject *scopeid = PyTuple_GetItem(tuple, 3);
				
#if PY_MAJOR_VERSION >= 3
				if(!PyUnicode_Check(host))
#else /* PY_MAJOR_VERSION >= 3 */
				if(!PyString_Check(host))
#endif /* PY_MAJOR_VERSION >= 3 */
				{
					PyErr_SetString(PyExc_AttributeError, "first item of sockaddr must be a string");
					return NULL;
				}
#if PY_MAJOR_VERSION >= 3
				if(!(PyUnicode_Check(port) || PyLong_Check(port)))
#else /* PY_MAJOR_VERSION >= 3 */
				if(!(PyString_Check(port) || PyInt_Check(port)))
#endif /* PY_MAJOR_VERSION >= 3 */
				{
					PyErr_SetString(PyExc_AttributeError, "second item of sockaddr must be an int or a string");
					return NULL;
				}
#if PY_MAJOR_VERSION >= 3
				if(!PyLong_Check(flowinfo))
#else /* PY_MAJOR_VERSION >= 3 */
				if(!PyInt_Check(flowinfo))
#endif /* PY_MAJOR_VERSION >= 3 */
				{
					PyErr_SetString(PyExc_AttributeError, "third item of sockaddr must be an int");
					return NULL;
				}
#if PY_MAJOR_VERSION >= 3
				if(!PyLong_Check(scopeid))
#else /* PY_MAJOR_VERSION >= 3 */
				if(!PyInt_Check(scopeid))
#endif /* PY_MAJOR_VERSION >= 3 */
				{
					PyErr_SetString(PyExc_AttributeError, "fourth item of sockaddr must be an int");
					return NULL;
				}
				
#if PY_MAJOR_VERSION >= 3
				PyObject *enchost;
				enchost = PyUnicode_AsUTF8String(host);
				char *addr = PyBytes_AsString(enchost);
#else /* PY_MAJOR_VERSION >= 3 */
				char *addr = PyString_AsString(host);
#endif /* PY_MAJOR_VERSION >= 3 */
				int rc = inet_pton(AF_INET6, addr, &s6.sin6_addr);
#if PY_MAJOR_VERSION >= 3
				Py_DECREF(enchost);
#endif /* PY_MAJOR_VERSION >= 3 */
				if(rc < 0)
				{
					PyErr_SetFromErrno(PyExc_AttributeError);
					return NULL;
				}
				
				int port_number = get_port(port);
				if(port_number < 0)
				{
					/* exception is set by get_port() */
					return NULL;
				}
				s6.sin6_port = htons(port_number);
				
#if PY_MAJOR_VERSION >= 3
				s6.sin6_flowinfo = htonl(PyLong_AsLong(flowinfo));
				s6.sin6_scope_id = htonl(PyLong_AsLong(scopeid));
#else /* PY_MAJOR_VERSION >= 3 */
				s6.sin6_flowinfo = htonl(PyInt_AsLong(flowinfo));
				s6.sin6_scope_id = htonl(PyInt_AsLong(scopeid));
#endif /* PY_MAJOR_VERSION >= 3 */
				
				sa = (struct sockaddr*)&s6;
				sa_size = sizeof(s6);
			}
			break;
		default: /* error */
			PyErr_SetString(PyExc_AttributeError, "tuple must be a valid sockaddr");
			return NULL;
			break;
	}
	
	//TODO: handle the 'flags' arg
	q->get_host = 1 /*get_host*/;
	q->get_serv = 1 /*get_serv*/;
	
	Query_store_asyncns(&q->query, self,
		asyncns_getnameinfo(self->asyncns,
		                    sa,
		                    sa_size,
		                    0, q->get_host, q->get_serv));
	
	asyncns_add_query(self, (Query*)q);
	return q;
}

static ResQuery *Asyncns_res_query(Asyncns *self, PyObject *args, PyObject *kwds)
{
	const char *dname = NULL;
	ns_class c;
	ns_type t;
	static char *kwlist[] = {"dname", "class", "type", NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "sii", kwlist, &dname, &c, &t))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
	ResQuery *q = (ResQuery *)PyObject_MyNew(ResQuery_type);
	
	Query_store_asyncns(&q->query, self,
		asyncns_res_query(self->asyncns, dname, c, t));
	asyncns_add_query(self, (Query*)q);
	return q;
}

static PyObject *Asyncns_wait(Asyncns *self, PyObject *args, PyObject *kwds)
{
	unsigned char block = 1;
	static char *kwlist[] = {"block", NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|b", kwlist, &block))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
	if(asyncns_wait(self->asyncns, (int)block) < 0)
	{
		PyErr_SetString(PyExc_Exception, "wait() failed");
		return NULL;
	}
	Py_RETURN_NONE;
}

static PyObject *Asyncns_get_next(Asyncns *self, PyObject *args, PyObject *kwds)
{
	asyncns_query_t *asyncns_query;
	static char *kwlist[] = {NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "", kwlist))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
	if(self->queries == NULL)
	{
		PyErr_SetString(PyExc_IndexError, "no reply available");
		return NULL;
	}
	asyncns_query = asyncns_getnext(self->asyncns);
	if(asyncns_query == NULL)
	{
		PyErr_SetString(PyExc_Exception, "no reply available");
		return NULL;
	}
	Query *query = _asyncns_get_query(self, asyncns_query);
	Py_XINCREF(query);
	return (PyObject*)query;
}

static PyObject * Asyncns_get_nqueries(Asyncns *self, PyObject *args, PyObject *kwds)
{
	static char *kwlist[] = {NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "", kwlist))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
#if PY_MAJOR_VERSION >= 3
	return PyLong_FromLong(asyncns_getnqueries(self->asyncns));
#else /* PY_MAJOR_VERSION >= 3 */
	return PyInt_FromLong(asyncns_getnqueries(self->asyncns));
#endif /* PY_MAJOR_VERSION >= 3 */
}

static PyObject * Asyncns_get_fd(Asyncns *self, PyObject *args, PyObject *kwds)
{
	static char *kwlist[] = {NULL};
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "", kwlist))
	{
		/* exception is set by PyArg_ParseTupleAndKeywords */
		return NULL;
	}
#if PY_MAJOR_VERSION >= 3
	return PyLong_FromLong(asyncns_fd(self->asyncns));
#else /* PY_MAJOR_VERSION >= 3 */
	return PyInt_FromLong(asyncns_fd(self->asyncns));
#endif /* PY_MAJOR_VERSION >= 3 */
}

static PyMethodDef Asyncns_methods[] =
{
	{ .ml_name = "getaddrinfo",
	  .ml_meth = (PyCFunction)Asyncns_getaddrinfo,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "getaddrinfo(host, proto, family = AF_UNSPEC, socktype = 0, proto = 0, flags = 0)\n"
	            "Issue a name to address query.\n"
	            "The arguments are compatible with the ones of the socket module.\n"
	            "The function returns a new AddrInfoQuery object." },
	{ .ml_name = "getnameinfo",
	  .ml_meth = (PyCFunction)Asyncns_getnameinfo,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "getnameinfo(sockaddr, flags = 0)\n"
	            "Issue an address to name query.\n"
	            "The arguments are compatible with the ones of the socket module.\n"
	            "The function returns a new NameInfoQuery object.\n" },
	{ .ml_name = "res_query",
	  .ml_meth = (PyCFunction)Asyncns_res_query,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "res_query(dname, class, type)\n"
	            "Issue an resolver query.\n"
	            "The arguments are compatible with the ones of libc's res_query(3).\n"
	            "The function returns a new ResQuery object.\n" },
	{ .ml_name = "wait",
	  .ml_meth = (PyCFunction)Asyncns_wait,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "wait(block = True)\n"
	            "Process pending events.\n"
	            "After this function is called you can get the next completed query\n"
	            "object(s) using get_next(). If block is True wait until at least\n"
	            "one response has been processed. If block is False, process\n"
	            "all pending responses and return." },
	{ .ml_name = "get_next",
	  .ml_meth = (PyCFunction)Asyncns_get_next,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "get_next()\n"
	            "Return the next completed query object.\n"
	            "If no query has been completed yet, return None.\n"
	            "Please note that you need to run wait() before this function\n"
	            "will return sensible data." },
	{ .ml_name = "get_nqueries",
	  .ml_meth = (PyCFunction)Asyncns_get_nqueries,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "get_nqueries()\n"
	            "Return the number of query objects (completed or not)." },
	{ .ml_name = "get_fd",
	  .ml_meth = (PyCFunction)Asyncns_get_fd,
	  .ml_flags = METH_VARARGS|METH_KEYWORDS,
	  .ml_doc = "get_fd()\n"
	            "Return the UNIX file descriptor to select() for readability on.\n"
	            "Use this function to integrate libasyncns with your custom main loop." },
	{ NULL, NULL, 0, NULL },
};

static PyMemberDef Asyncns_members[] =
{
	{ NULL, 0, 0, 0, NULL },
};

PyTypeObject Asyncns_type =
{
#if PY_MAJOR_VERSION >= 3
	PyVarObject_HEAD_INIT(NULL,0)
#else /* PY_MAJOR_VERSION >= 3 */
	PyObject_HEAD_INIT(NULL)
#endif /* PY_MAJOR_VERSION >= 3 */

	.tp_name = "libasyncns.Asyncns",
	.tp_doc  = "The session class.\n"
	           "Asyncns(n_proc = 5)",

	.tp_new     = (newfunc)   Asyncns_new,
	.tp_init    = (initproc)  Asyncns_init,
	.tp_dealloc = (destructor)Asyncns_dealloc,

	.tp_methods = Asyncns_methods,
	.tp_members = Asyncns_members,

	.tp_flags     = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
	.tp_basicsize = sizeof(Asyncns),
};

void initAsyncns(PyObject *libasyncns_module)
{
	PyType_Ready(&Asyncns_type);
	Py_INCREF(&Asyncns_type);
	PyModule_AddObject(libasyncns_module, "Asyncns", (PyObject*)&Asyncns_type);
}
