/*
Copyright (C) 2008  Damien Thebault

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <Python.h>
#include <structmember.h>
#include <arpa/nameser.h>

#include "libasyncns.h"
#include "asyncns.h"
#include "query.h"
#include "addrinfoquery.h"
#include "nameinfoquery.h"
#include "resquery.h"

static PyMethodDef libasyncns_methods[] =
{
	{ NULL, NULL, 0, NULL, },
};

static PyObject *socket_gaierror;

static void init_gaierror(void)
{
	PyObject *socket_module = PyImport_ImportModule("_socket");
	if(socket_module == NULL)
	{
		PyErr_SetString(PyExc_ImportError, "could not import socket");
		return;
	}
	
	socket_gaierror = PyObject_GetAttrString(socket_module, "gaierror");
	Py_DECREF(socket_module);
	if(socket_gaierror == NULL)
	{
		PyErr_SetString(PyExc_AttributeError, "can't find attribute \"gaierror\" in the socket module");
		return;
	}
}

#if PY_MAJOR_VERSION >= 3
#undef PyModuleDef_HEAD_INIT
#define PyModuleDef_HEAD_INIT {PyObject_HEAD_INIT(NULL)NULL,0,NULL}
static struct PyModuleDef libasyncns_moduledef =
{
	PyModuleDef_HEAD_INIT,
	"libasyncns",
	NULL,
	0,
	libasyncns_methods,
	NULL,
	NULL,
	NULL,
	NULL,
};
#else /* PY_MAJOR_VERSION >= 3 */
static char libasyncns_name_raw[] = "libasyncns";
#endif /* PY_MAJOR_VERSION >= 3 */

#if PY_MAJOR_VERSION >= 3
PyObject * PyInit_libasyncns(void)
#else /* PY_MAJOR_VERSION >= 3 */
PyMODINIT_FUNC initlibasyncns(void)
#endif /* PY_MAJOR_VERSION >= 3 */
{

#if PY_MAJOR_VERSION >= 3
	PyObject *libasyncns_module = PyModule_Create(&libasyncns_moduledef);
#else /* PY_MAJOR_VERSION >= 3 */
	PyObject *libasyncns_module = Py_InitModule(libasyncns_name_raw,
	                                            libasyncns_methods);
#endif /* PY_MAJOR_VERSION >= 3 */
	
#define ADDNSCONST(const) PyModule_AddIntConstant(libasyncns_module, #const, const)
	ADDNSCONST(ns_t_invalid);
	ADDNSCONST(ns_t_a);
	ADDNSCONST(ns_t_ns);
	ADDNSCONST(ns_t_md);
	ADDNSCONST(ns_t_mf);
	ADDNSCONST(ns_t_cname);
	ADDNSCONST(ns_t_soa);
	ADDNSCONST(ns_t_mb);
	ADDNSCONST(ns_t_mg);
	ADDNSCONST(ns_t_mr);
	ADDNSCONST(ns_t_null);
	ADDNSCONST(ns_t_wks);
	ADDNSCONST(ns_t_ptr);
	ADDNSCONST(ns_t_hinfo);
	ADDNSCONST(ns_t_minfo);
	ADDNSCONST(ns_t_mx);
	ADDNSCONST(ns_t_txt);
	ADDNSCONST(ns_t_rp);
	ADDNSCONST(ns_t_afsdb);
	ADDNSCONST(ns_t_x25);
	ADDNSCONST(ns_t_isdn);
	ADDNSCONST(ns_t_rt);
	ADDNSCONST(ns_t_nsap);
	ADDNSCONST(ns_t_nsap_ptr);
	ADDNSCONST(ns_t_sig);
	ADDNSCONST(ns_t_key);
	ADDNSCONST(ns_t_px);
	ADDNSCONST(ns_t_gpos);
	ADDNSCONST(ns_t_aaaa);
	ADDNSCONST(ns_t_loc);
	ADDNSCONST(ns_t_nxt);
	ADDNSCONST(ns_t_eid);
	ADDNSCONST(ns_t_nimloc);
	ADDNSCONST(ns_t_srv);
	ADDNSCONST(ns_t_atma);
	ADDNSCONST(ns_t_naptr);
	ADDNSCONST(ns_t_kx);
	ADDNSCONST(ns_t_cert);
	ADDNSCONST(ns_t_a6);
	ADDNSCONST(ns_t_dname);
	ADDNSCONST(ns_t_sink);
	ADDNSCONST(ns_t_opt);
	ADDNSCONST(ns_t_tsig);
	ADDNSCONST(ns_t_ixfr);
	ADDNSCONST(ns_t_axfr);
	ADDNSCONST(ns_t_mailb);
	ADDNSCONST(ns_t_maila);
	ADDNSCONST(ns_t_any);
	ADDNSCONST(ns_t_zxfr);
	ADDNSCONST(ns_t_max);
	
	ADDNSCONST(ns_c_invalid);
	ADDNSCONST(ns_c_in);
	ADDNSCONST(ns_c_2);
	ADDNSCONST(ns_c_chaos);
	ADDNSCONST(ns_c_hs);
	ADDNSCONST(ns_c_none);
	ADDNSCONST(ns_c_any);
	ADDNSCONST(ns_c_max);
	
	init_gaierror();
	
	initQuery(libasyncns_module);
	initAsyncns(libasyncns_module);
	initAddrInfoQuery(libasyncns_module);
	initNameInfoQuery(libasyncns_module);
	initResQuery(libasyncns_module);
#if PY_MAJOR_VERSION >= 3
	return libasyncns_module;
#endif /* PY_MAJOR_VERSION >= 3 */
}

void libasyncns_set_gaierror(int error)
{
	PyObject *value = Py_BuildValue("(is)", error, gai_strerror(error));
	if (value != NULL)
	{
		if(socket_gaierror != NULL)
		{
			PyErr_SetObject(socket_gaierror, value);
		}
		else
		{
			PyErr_SetObject(PyExc_Exception, value);
		}
		Py_DECREF(value);
	}
}

