/*
Copyright (C) 2008  Damien Thebault

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "libasyncns.h"
#include "nameinfoquery.h"
#include "query.h"

#include <structmember.h>

static int NameInfoQuery_init(NameInfoQuery *self, PyObject *args, PyObject *kwds)
{
	Query_type.tp_init((PyObject*)self, args, kwds);
	self->get_host = 0;
	self->get_serv = 0;
	/* exception is set by Query.__init__() */
	return -1;
}

static PyObject *NameInfoQuery_get_done(NameInfoQuery *self)
{
	PyObject *out = NULL;
	char host[NI_MAXHOST];
	char serv[NI_MAXSERV];
	int r;
	if(self->query.result != NULL)
	{
		out = self->query.result;
		Py_INCREF(out);
		goto get_done_end;
	}
	r = asyncns_getnameinfo_done(self->query.asyncns->asyncns,
				self->query.asyncns_query,
				host, sizeof(host), serv, sizeof(serv));
	asyncns_del_query(self->query.asyncns, (Query*)self);
	if(r != 0)
	{
		libasyncns_set_gaierror(r);
		out = NULL;
		goto get_done_end;
	}
	out = PyTuple_New(2);
	if(self->get_host)
	{
#if PY_MAJOR_VERSION >= 3
		PyTuple_SetItem(out, 0, PyUnicode_FromString(host));
#else /* PY_MAJOR_VERSION >= 3 */
		PyTuple_SetItem(out, 0, PyString_FromString(host));
#endif /* PY_MAJOR_VERSION >= 3 */
	}
	if(self->get_serv)
	{
#if PY_MAJOR_VERSION >= 3
		PyTuple_SetItem(out, 1, PyUnicode_FromString(serv));
#else /* PY_MAJOR_VERSION >= 3 */
		PyTuple_SetItem(out, 1, PyString_FromString(serv));
#endif /* PY_MAJOR_VERSION >= 3 */
	}
	Query_store_result(&(self->query), out);
get_done_end:
	self->query.asyncns_query = NULL;
	return out;
}

static PyMethodDef NameInfoQuery_methods[] =
{
	{ .ml_name = "get_done",
	  .ml_meth = (PyCFunction)NameInfoQuery_get_done,
	  .ml_flags = METH_NOARGS,
	  .ml_doc = "get_done()\n"
	            "Return the result of the query.\n"
	            "The result is a tuple of 2 items.\n"
	            "The first item is the host name, "
	            "the second one is the service name.\n" },
	{ NULL, NULL, 0, NULL },
};

static PyMemberDef NameInfoQuery_members[] =
{
	{ NULL, 0, 0, 0, NULL },
};


PyTypeObject NameInfoQuery_type =
{
#if PY_MAJOR_VERSION >= 3
	PyVarObject_HEAD_INIT(NULL,0)
#else /* PY_MAJOR_VERSION >= 3 */
	PyObject_HEAD_INIT(NULL)
#endif /* PY_MAJOR_VERSION >= 3 */

	.tp_name = "libasyncns.NameInfoQuery",
	.tp_doc  = "The class for getnameinfo() calls.",

	.tp_init    = (initproc)  NameInfoQuery_init, /* no init, see Asyncns */

	.tp_methods = NameInfoQuery_methods,
	.tp_members = NameInfoQuery_members,

	.tp_flags     = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
	.tp_basicsize = sizeof(NameInfoQuery),
};

void initNameInfoQuery(PyObject *libasyncns_module)
{
	NameInfoQuery_type.tp_base = &Query_type; /* windows workaround */
	PyType_Ready(&NameInfoQuery_type);
	Py_INCREF(&NameInfoQuery_type);
	PyModule_AddObject(libasyncns_module, "NameInfoQuery", (PyObject*)&NameInfoQuery_type);
}
