/*
Copyright (C) 2008  Damien Thebault

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "query.h"
#include <structmember.h>

static Query *Query_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
	Query *self = (Query*)type->tp_alloc(type, 0);
	return self;
}

void Query_dealloc(Query *self)
{
	if(self->asyncns_query && self->asyncns && self->asyncns->asyncns &&
           !asyncns_isdone(self->asyncns->asyncns, self->asyncns_query))
		asyncns_cancel(self->asyncns->asyncns, self->asyncns_query);
	if(self->asyncns != NULL)
		asyncns_del_query(self->asyncns, self);
	self->asyncns_query = NULL;
	Py_XDECREF(self->asyncns);
	self->asyncns = NULL;
	Py_XDECREF(self->result);
	self->result = NULL;
}

static int Query_init(Query *self, PyObject *args, PyObject *kwds)
{
	self->asyncns = NULL;
	self->asyncns_query = NULL;
	self->result = NULL;
	self->userdata = NULL;
	PyErr_SetString(PyExc_RuntimeError, "this class is only available using Asyncns' functions");
	return -1;
}

void Query_store_asyncns(Query *self, Asyncns *asyncns, asyncns_query_t *asyncns_query)
{
	self->asyncns = asyncns;
	Py_INCREF(asyncns);
	self->asyncns_query = asyncns_query;
}

void Query_store_result(Query *self, PyObject *result)
{
	self->result = result;
	Py_INCREF(result);
}

PyObject *Query_is_done(Query *self)
{
	int done = 0;
	if(self->result != NULL)
	{
		done = 1;
	}
	else
	{
		done = asyncns_isdone(self->asyncns->asyncns, self->asyncns_query);
	}
	return PyBool_FromLong(done);
}

PyObject *Query_cancel(Query *self)
{
	asyncns_cancel(self->asyncns->asyncns, self->asyncns_query);
	Py_RETURN_NONE;
}

static PyMethodDef Query_methods[] =
{
	{ .ml_name = "is_done",
	  .ml_meth = (PyCFunction)Query_is_done,
	  .ml_flags = METH_NOARGS,
	  .ml_doc = "is_done()\n"
	            "Tells if the request has been completed." },
	{ .ml_name = "cancel",
	  .ml_meth = (PyCFunction)Query_cancel,
	  .ml_flags = METH_NOARGS,
	  .ml_doc = "cancel()\n"
	            "Cancels the query." },
	{ NULL, NULL, 0, NULL },
};

static PyMemberDef Query_members[] =
{
	{ "userdata", T_OBJECT,
	  offsetof(Query, userdata), 0,
	  "Attribute available to store user data." },
	{ NULL, 0, 0, 0, NULL },
};

PyTypeObject Query_type =
{
#if PY_MAJOR_VERSION >= 3
	PyVarObject_HEAD_INIT(NULL,0)
#else /* PY_MAJOR_VERSION >= 3 */
	PyObject_HEAD_INIT(NULL)
#endif /* PY_MAJOR_VERSION >= 3 */

	.tp_name = "libasyncns.Query",
	.tp_doc  = "A generic query class.",

	.tp_new     = (newfunc)   Query_new,
	.tp_dealloc = (destructor)Query_dealloc,
	.tp_init    = (initproc)  Query_init, /* no init, see Asyncns */

	.tp_methods = Query_methods,
	.tp_members = Query_members,

	.tp_flags     = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
	.tp_basicsize = sizeof(Query),
};

void initQuery(PyObject *libasyncns_module)
{
	PyType_Ready(&Query_type);
	Py_INCREF(&Query_type);
	PyModule_AddObject(libasyncns_module, "Query", (PyObject*)&Query_type);
}
