/*
Copyright (C) 2008  Damien Thebault

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "libasyncns.h"
#include "resquery.h"
#include "query.h"

#include <structmember.h>
#include <arpa/nameser.h>
#include <resolv.h>

static int ResQuery_init(ResQuery *self, PyObject *args, PyObject *kwds)
{
	Query_type.tp_init((PyObject*)self, args, kwds);
	/* exception is set by Query.__init__() */
	return -1;
}

static PyObject *ResQuery_get_done(ResQuery *self)
{
	PyObject *out = NULL;
	unsigned char *srv = NULL;
	int r;
	if(self->query.result != NULL)
	{
		out = self->query.result;
		Py_INCREF(out);
		goto get_done_end;
	}
	r = asyncns_res_done(self->query.asyncns->asyncns, self->query.asyncns_query, &srv);
	asyncns_del_query(self->query.asyncns, (Query*)self);
	if(r < 0)
	{
		libasyncns_set_gaierror(r);
		out = NULL;
		goto get_done_end;
	}
	else if(srv == NULL)
	{
		PyErr_SetString(PyExc_Exception, "no valid reply");
		out = NULL;
		goto get_done_end;
	}
	out = PyList_New(0);
	int qdcount, ancount, len;
	const unsigned char *pos = srv+sizeof(HEADER);
	unsigned char *end = srv+r;
	HEADER *head = (HEADER*)srv;
	char name[256];
	qdcount = ntohs(head->qdcount);
	ancount = ntohs(head->ancount);
	/* Ignore the questions */
	while(qdcount-- > 0 && (len = dn_expand(srv, end, pos, name, 255)) >= 0)
	{
		assert(len >= 0);
		pos += len + QFIXEDSZ;
	}
	/* Parse the answers */
	while(ancount-- > 0 && (len = dn_expand(srv, end, pos, name, 255)) >= 0)
	{
		uint16_t type, class, length, pref, weight, port;
		uint32_t ttl;
		/* Ignore the initial string */
		assert(len >= 0);
		pos += len;
		GETSHORT(type, pos);
		GETSHORT(class, pos);
		GETLONG(ttl, pos);
		GETSHORT(length, pos);
		GETSHORT(pref, pos);
		GETSHORT(weight, pos);
		GETSHORT(port, pos);
		len = dn_expand(srv, end, pos, name, 255);
		pos += len;
		PyObject *o = PyDict_New();
#if PY_MAJOR_VERSION >= 3
		PyDict_SetItemString(o, "type",   PyLong_FromLong(type));
		PyDict_SetItemString(o, "class",  PyLong_FromLong(class));
		PyDict_SetItemString(o, "ttl",    PyLong_FromLong(ttl));
		PyDict_SetItemString(o, "length", PyLong_FromLong(length));
		PyDict_SetItemString(o, "pref",   PyLong_FromLong(pref));
		PyDict_SetItemString(o, "weight", PyLong_FromLong(weight));
		PyDict_SetItemString(o, "port",   PyLong_FromLong(port));
		PyDict_SetItemString(o, "host",   PyUnicode_FromString(name));
#else /* PY_MAJOR_VERSION >= 3 */
		PyDict_SetItemString(o, "type",   PyInt_FromLong(type));
		PyDict_SetItemString(o, "class",  PyInt_FromLong(class));
		PyDict_SetItemString(o, "ttl",    PyInt_FromLong(ttl));
		PyDict_SetItemString(o, "length", PyInt_FromLong(length));
		PyDict_SetItemString(o, "pref",   PyInt_FromLong(pref));
		PyDict_SetItemString(o, "weight", PyInt_FromLong(weight));
		PyDict_SetItemString(o, "port",   PyInt_FromLong(port));
		PyDict_SetItemString(o, "host",   PyString_FromString(name));
#endif /* PY_MAJOR_VERSION >= 3 */
		
		PyList_Append(out, o);
	}
	if(srv != NULL)
		asyncns_freeanswer(srv);
	Query_store_result(&(self->query), out);
get_done_end:
	self->query.asyncns_query = NULL;
	return out;
}

static PyMethodDef ResQuery_methods[] =
{
	{ .ml_name = "get_done",
	  .ml_meth = (PyCFunction)ResQuery_get_done,
	  .ml_flags = METH_NOARGS,
	  .ml_doc = "get_done()\n"
	            "Return the result of the query.\n"
	            "The result is a list of dictionaries with the following keys:\n"
	            "'weight', "
	            "'pref', "
	            "'host', "
	            "'port', "
	            "'length', "
	            "'ttl', "
	            "'type', "
	            "'class'" },
	{ NULL, NULL, 0, NULL },
};

static PyMemberDef ResQuery_members[] =
{
	{ NULL, 0, 0, 0, NULL },
};


PyTypeObject ResQuery_type =
{
#if PY_MAJOR_VERSION >= 3
	PyVarObject_HEAD_INIT(NULL,0)
#else /* PY_MAJOR_VERSION >= 3 */
	PyObject_HEAD_INIT(NULL)
#endif /* PY_MAJOR_VERSION >= 3 */

	.tp_name = "libasyncns.ResQuery",
	.tp_doc  = "The class for res_query() calls.",

	.tp_init    = (initproc)  ResQuery_init, /* no init, see Asyncns */

	.tp_methods = ResQuery_methods,
	.tp_members = ResQuery_members,

	.tp_flags     = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
	.tp_basicsize = sizeof(ResQuery),
};

void initResQuery(PyObject *libasyncns_module)
{
	ResQuery_type.tp_base = &Query_type; /* windows workaround */
	PyType_Ready(&ResQuery_type);
	Py_INCREF(&ResQuery_type);
	PyModule_AddObject(libasyncns_module, "ResQuery", (PyObject*)&ResQuery_type);
}
